@@ -1,7 +1,7 @@
 #!/usr/bin/perl -w
 
 ###############################################################################
-# cvsplot: Copyright (c) 2001, 2002 David Sitsky.  All rights reserved.
+# cvsplot: Copyright (c) 2001, 2002, 2003 David Sitsky.  All rights reserved.
 #
 # cvsplot is a perl script which is used to extract information from CVS and
 # plots the total number of lines and number of files in a selected file set
@@ -13,8 +13,12 @@
 # This program is free software; you can redistribute it and modify it under
 # the terms of the GPL.
 
+use Config;
 use Cwd;
 use Date::Manip;
+use String::ShellQuote qw(shell_quote);
+use Symbol;
+use IPC::Open3;
 
 # Whether debugging is enabled or not.
 $debug = 0;
@@ -78,6 +82,13 @@ $filedata = "";
 # Flag to indicate if gnuplot is to be used.
 $use_gnuplot = 0;
 
+# The location of the gnuplot binary.  It is assumed to be in the
+# path, but may be over-ridden on the command-line with the
+# -gnuplot argument.  For Window platforms, this is useful, since
+# the name of the binary is not "gnuplot", but often "wgnupl32"
+# or "pgnuplot".
+$gnuplot = "gnuplot";
+
 # The gnuplot output filename to write the output of the line data.
 $gnuplot_linedata = "";
 
@@ -92,12 +103,40 @@ $gnuplot_setterm = "png color";
 # aspect of the plotting command, such as the format of the x values.
 $gnuplot_command = "";
 
+# Flag to indicate if lines are to be used in the final plot, rather
+# than discrete points.  For large plots, this can sometimes improve
+# readability.
+$use_linestyle = 0;
+
+# Determine if this process is running under Windows.
+$osname = $Config{'osname'};
+$windows = (defined $osname && $osname eq "MSWin32") ? 1 : 0;
+
 process_command_line_arguments();
 get_cvs_statistics();
 analyse_statistics();
 generate_data_files();
 generate_plots();
 
+###############################################################################
+# Utility method for quoting an argument for a shell command.  ShellQuote
+# is good for UNIX boxes, but doesn't work for DOS platforms as it uses
+# single quotes, while DOS needs double quotes.  Its a shame shell_quote
+# isn't cross-platform.
+sub quote
+{
+    my ($arg) = @_;
+
+    if ($windows)
+    {
+        return "\"$arg\"";
+    }
+    else
+    {
+        shell_quote($arg);
+    }
+}
+
 ###############################################################################
 # Check whether the supplied file is to be examined or not depending on what
 # the user set for the -include and -exclude options.  Return true if the
@@ -135,6 +174,18 @@ sub include_file
 #
 sub get_cvs_statistics
 {
+    if ($debug && defined $osname)
+    {
+        print "Platform is $osname\n";
+    }
+
+    # Explicitly set the timezone for window platforms, so that DateManip
+    # works.
+    if ($windows)
+    {
+        $ENV{TZ} = "C";
+    }
+
     my $working_file = "";
     my $relative_working_file = "";
     my $working_cvsdir = "";
@@ -145,7 +196,7 @@ sub get_cvs_statistics
     if ($rlog_module eq "")
     {
 	$saved_cwd = cwd();
-	chdir $cvsdir;
+	chdir $cvsdir || die "Failed to change to directory \"$cvsdir\": $!";
     }
     else
     {
@@ -153,6 +204,10 @@ sub get_cvs_statistics
 	$cvsdir =~ /([^:]+)$/;
 	$working_cvsdir = $1;
 	print "Got working_cvsdir as $working_cvsdir\n" if $debug;
+
+	# Since this is used in a regexp below, need to make sure DOS pathnames
+	# are correctly matched against.
+	$working_cvsdir =~ s/\\/\\\\/g;
     }
 
     # Flag to indicate what the state is when parsing the output from cvs log.
@@ -162,8 +217,9 @@ sub get_cvs_statistics
 
     # Build up the command string appropriately, depending on what options
     # have been set.
-    my $command =
-	($rlog_module ne "") ? "cvs -d $cvsdir rlog $rlog_module" : "cvs log";
+    my $command = ($rlog_module eq '') ? 'cvs log' :
+	sprintf('cvs -d %s rlog %s', quote($cvsdir), quote($rlog_module));
+
     print "Executing \"$command\"\n" if $debug;
 
     open (CVSLOG, "$command |") || die "Couldn't execute \"$command\"";
@@ -206,6 +262,14 @@ sub get_cvs_statistics
 	    {
 		$relative_working_file = $1;
 	    }
+	    # Handle repositories working off an explicit numbering scheme,
+	    # such as 8.1.  Only do this if the user hasn't specified an
+	    # explicit branch to gather statistics over.  In most cases,
+	    # the result will still be 1, but this handles the stranger
+	    # repositories out there.
+	    elsif ($branch_tag eq "" && /^head: (\d+)\./) {
+		$file_branch_number{$working_file} = $1;       
+	    }
 	    # If we are collecting statistics on a branch, determine the magic
 	    # branch number for this file.
 	    elsif ( (! defined $file_branch_number{$working_file}) &&
@@ -282,7 +346,7 @@ sub get_cvs_statistics
 		}
 		    
 		$_ = <CVSLOG>;		# Read the "date" line.
-		if (/^date: (\d\d\d\d\/\d\d\/\d\d \d\d:\d\d:\d\d); .* state: (.*);.*lines: \+(\d+) \-(\d+)$/)
+		if (/^date: (\d\d\d\d\/\d\d\/\d\d \d\d:\d\d:\d\d); .* state: (.*);.*lines: \+(\d+) \-(\d+).*$/)
 		{
 		    # Note for some CVS clients, state dead is presented in
 		    # this this way, as the following pattern.
@@ -304,7 +368,7 @@ sub get_cvs_statistics
 			$revision_stats{$date}{$working_file} = $revision;
 		    }
 	        }
-		elsif (/^date: (\d\d\d\d\/\d\d\/\d\d \d\d:\d\d:\d\d); .* state: dead;$/)
+		elsif (/^date: (\d\d\d\d\/\d\d\/\d\d \d\d:\d\d:\d\d); .* state: dead;.*$/)
 		{
 		    # File has been removed.
 		    $date = $1;
@@ -319,33 +383,69 @@ sub get_cvs_statistics
 			$revision_stats{$date}{$working_file} = $revision;
 		    }
 		}
-		elsif (/^date: (\d\d\d\d\/\d\d\/\d\d \d\d:\d\d:\d\d); .* state: Exp;$/)
+		elsif (/^date: (\d\d\d\d\/\d\d\/\d\d \d\d:\d\d:\d\d); .* state: ([^;]*);.*$/)
 		{
 		    $date = $1;
+		    my $state = $2;
 
 		    # Unfortunately, cvs log doesn't indicate the number of
 		    # lines an initial revision is created with, so find this
-		    # out using the following cvs command.
+		    # out using the following cvs command.  Note the regexp
+		    # below has an optional drive delimeter to support DOS
+		    # installations.
 		    my $lccmd = "";
 		    if ($rlog_module ne "")
 		    {
-			$working_file =~ /^${working_cvsdir}\/(.*)$/;
-			$lccmd = "cvs -d $cvsdir co -r $revision -p \"$1\"";
+			print "Working cvsdir is: $working_cvsdir working file: $working_file\n" if $debug;
+
+			# For DOS-based repositories, the filename may contain
+			# a drive letter.  Also need to be flexible with the
+			# pathname separator.
+			if (! ($working_file =~ /^([A-z]:)?${working_cvsdir}[\/\\](.*)$/))
+			{
+			    print STDERR "-cvsdir argument $working_cvsdir doesn't match ";
+			    print STDERR "repository filename prefix $working_file\n";
+			    print STDERR "Please correct your -cvsdir argument and try again\n";
+			    exit 1;
+			}
+			$lccmd = sprintf('cvs -d %s co -r %s -p %s',
+					 quote($cvsdir),
+					 quote($revision),
+					 quote($2));
 		    }
 		    else
 		    {
-			$lccmd = "cvs update -r $revision -p \"$relative_working_file\"";
+			$lccmd = sprintf('cvs update -r %s -p %s',
+					 quote($revision),
+					 quote($relative_working_file));
 		    }
 		    print "Executing $lccmd\n" if $debug;
-		    $number_lines = `$lccmd 2>/dev/null | wc -l`;
-		    chop $number_lines;
-		    $number_lines =~ s/ //g;
+		    
+		    my $WTR = gensym();
+		    my $RDR = gensym();
+		    my $ERR = gensym();
+		    my $pid = open3($WTR, $RDR, $ERR, $lccmd);
+		    for ($number_lines = 0; defined <$RDR>; $number_lines++) {}
+		    close ($RDR);
+		    my $error_string = "";
+		    while (<$ERR>)
+		    {
+			$error_string .= $_;
+		    }
+		    waitpid $pid, 0;
+		    
+		    if ($?) {
+			print "CVS command failed: \"$lccmd\" status $?\n";
+			print "$error_string\n";
+			exit 1;
+		    }
+		    
 		    print "$working_file 1.1 = $number_lines lines\n" if $debug;
-
+		    
 		    $file_version_delta{$working_file}{$revision} =
 			$number_lines;
-		    $file_version_state{$working_file}{$revision} = "Exp";
-
+		    $file_version_state{$working_file}{$revision} = $state;
+		    
 		    if ($file_on_branch)
 		    {
 			$line_stats{$date}{$working_file} += $number_lines;
@@ -366,6 +466,7 @@ sub get_cvs_statistics
 	    }
 	}
     }
+    close(CVSLOG);
 
     # Go back to the original directory if we aren't using the -rlog option.
     if ($rlog_module eq "")
@@ -574,30 +675,82 @@ sub generate_data_files
 ###############################################################################
 # Generate the gnuplot data files into the specified locations.
 #
+# useful reference : http://www.cs.uni.edu/Help/gnuplot/TOC.html
+#
 sub generate_plots
 {
+    my $linestyle_command = "";
+    if ($use_linestyle)
+    {
+	$linestyle_command = "set data style lines";
+    }
+
     if ($use_gnuplot)
     {
-	# Generate a gnuplot command script to build the necessary images.
-	$command = <<EOF;
+	# Generate the gnuplot command scripts to build the necessary images.
+
+	if ($gnuplot_linefiledata)
+	{
+	    # Produce a combined plot if the user has specified to build one.
+	    $command = <<EOF;
+
+set xdata time
+set timefmt '%Y/%m/%d %H:%M:%S'
+set format x '%m/%y'
+set xlabel 'Date'
+set title 'CVS Line/File Statistics'
+set key left top
+set ylabel 'Number of lines'
+set yrange [0:*]
+set ytics
+set y2label 'Number of files'
+set y2range [0:*]
+set y2tics
+set terminal $gnuplot_setterm
+set output '$gnuplot_linefiledata'
+$linestyle_command
+$gnuplot_command
+plot '$linedata' using 1:3 axes x1y1 title 'Lines',\\
+     '$filedata' using 1:3 axes x1y2 title 'Files'
+EOF
+
+            # Pipe this command into gnuplot.
+            print "Issuing command:\n\n$command\n\n to gnuplot: $gnuplot\n" if $debug;
+	    open(GNUPLOT, "| $gnuplot") || die "Failed to start gnuplot: $!";
+            print GNUPLOT $command;
+            close GNUPLOT;
+        }
+
+        # Produce two separate plots, one for the file data, one for the
+        # line data.
+        $command = <<EOF;
 
 set xdata time
 set timefmt '%Y/%m/%d %H:%M:%S'
 set format x '%m/%y'
 set xlabel 'Date'
 set ylabel 'Number of lines'
+set title 'CVS Line Statistics'
+set yrange [0:*]
 set nokey
 set terminal $gnuplot_setterm
 set output '$gnuplot_linedata'
+$linestyle_command
 $gnuplot_command
 plot '$linedata' using 1:3
+set title 'CVS Line Statistics'
 set ylabel 'Number of files'
+set yrange [0:*]
 set output '$gnuplot_filedata'
+$gnuplot_command
 plot '$filedata' using 1:3
 EOF
 
         # Pipe this command into gnuplot.
-	`echo "$command" | gnuplot`;
+        print "Issuing command:\n\n$command\n\n to gnuplot: $gnuplot\n" if $debug;
+	open(GNUPLOT, "| $gnuplot") || die "Failed to start gnuplot: $!";
+        print GNUPLOT $command;
+        close GNUPLOT;
     }
 }
 
@@ -613,6 +766,11 @@ sub process_command_line_arguments
 	    $debug = 1;
 	    $i++;
 	}
+	elsif ($ARGV[$i] eq "-linestyle")
+	{
+	    $use_linestyle = 1;
+	    $i++;
+	}
 	elsif ($ARGV[$i] eq "-include")
 	{
 	    $pattern_include[++$#pattern_include] = 1;
@@ -660,6 +818,12 @@ sub process_command_line_arguments
 	    $filedata = $ARGV[$i+1];
 	    $i += 2;
 	}
+	elsif ($ARGV[$i] eq "-gnuplot")
+	{
+	    $use_gnuplot = 1;
+	    $gnuplot = $ARGV[$i+1];
+	    $i += 2;
+	}
 	elsif ($ARGV[$i] eq "-gnuplotlinedata")
 	{
 	    $use_gnuplot = 1;
@@ -672,6 +836,12 @@ sub process_command_line_arguments
 	    $gnuplot_filedata = $ARGV[$i+1];
 	    $i += 2;
 	}
+	elsif ($ARGV[$i] eq "-gnuplotlinefiledata")
+	{
+	    $use_gnuplot = 1;
+	    $gnuplot_linefiledata = $ARGV[$i+1];
+	    $i += 2;
+	}
 	elsif ($ARGV[$i] eq "-gnuplotsetterm")
 	{
 	    $use_gnuplot = 1;
@@ -723,8 +893,8 @@ sub process_command_line_arguments
     }
 
 
-    # The line and file gnuplot options need to be set if a gnuplot option is
-    # used.
+    # The line and file (or linefile) gnuplot options need to be set if a
+    # gnuplot option is used.
     if ($use_gnuplot)
     {
 	if ($gnuplot_linedata eq "" || $gnuplot_filedata eq "")
@@ -747,11 +917,26 @@ sub process_command_line_arguments
     # If the -rlog option has been specified, need to make sure that the
     # CVS version install is >= 1.11.1, as it is not supported in earlier
     # versions.
-    if ($rlog_module ne "" &&
-	system("cvs rlog 2>&1 | grep deprecated > /dev/null") == 0)
+    if ($rlog_module ne "")
     {
-	print "error: -rlog option requires CVS version >= 1.11.1\n\n";
-	exit 1;
+	my $WTR = gensym();
+	my $RDR = gensym();
+	my $ERR = gensym();
+	my $pid = open3($WTR, $RDR, $ERR, "cvs rlog");
+	my $deprecated_found = 0;
+        while (<$ERR>)
+        {
+            $deprecated_found = 1 if (/deprecated/);
+        }
+        close $WTR;
+        close $RDR;
+        close $ERR;
+        waitpid $pid, 0;
+	if ($deprecated_found)
+        {
+	    print "error: -rlog option requires CVS version >= 1.11.1\n\n";
+	    exit 1;
+        }
     }
 
     # Open the specified output files.
@@ -767,17 +952,19 @@ sub process_command_line_arguments
 #
 sub usage
 {
-    print "cvsplot version 1.6.1 - ";
+    print "cvsplot version 1.6.5 - ";
     print "Copyright David Sitsky: sits\@users.sourceforge.net\n\n";
     print "cvsplot collects statistics from CVS controlled files.\n\n";
-    print "usage: cvsplot.pl -cvsdir <dir> [-rlog <module>]\n";
-    print "                  -linedata <file> -filedata <file>\n";
-    print "                  [-branch <branch name>] [-start <date>] [-end <date>]\n";
-    print "                  {-include <regexp> | -exclude <regexp>}\n";
-    print "                  [-gnuplotlinedata <gnuplot output file for line data>\n";
-    print "                   -gnuplotfiledata <gnuplot output file for file data>\n";
-    print "                   [-gnuplotsetterm <gnuplot set terminal expression>]\n";
-    print "                   [-gnuplotcommand <general gnuplot command>]]\n\n";
+    print "usage:\n";
+    print "cvsplot.pl -cvsdir <dir> [-rlog <module>] -linedata <file> -filedata <file>\n";
+    print "           [-branch <branch name>] [-start <date>] [-end <date>]\n";
+    print "           {-include <regexp> | -exclude <regexp>} [-linesstyle]\n";
+    print "           [-gnuplotlinedata <output file for line data>\n";
+    print "            -gnuplotfiledata <output file for file data>\n";
+    print "            [-gnuplot <path to gnuplot binary>]\n";
+    print "            [-gnuplotlinefiledata <output file combined line and file data>]\n";
+    print "            [-gnuplotsetterm <gnuplot set terminal expression>]\n";
+    print "            [-gnuplotcommand <general gnuplot command>]]\n\n";
     print "See http://cvsplot.sourceforge.net for updates.\n";
     exit 1;
 }
